#!/usr/bin/python3
# -*- coding: utf-8 -*-
# waveform.py
# D. Gibbon
# 2019-04-20
#==================================================================

try:
	print("Importing sys, re.")
	import sys, re
	ver = sys.version[0]
	if ver == "3":
		print("Using Python3.")
		from tkinter import Tk
		from tkinter.filedialog import askopenfilename
	elif ver == "2":
		print("Using Python2. But Python 3 is required by most ProsodySuite apps.")
		from Tkinter import Tk
		from tkFileDialog import askopenfilename
	else:
		print("Unsupported Python version %s"%ver); exit()
	print("Importing from numpy, matplotlib, scipy.")
	import numpy as np
	import matplotlib.pyplot as plt
	import scipy.io.wavfile as wave
	from scipy.signal import medfilt, butter, lfilter, hilbert
	import pyaudio
	p = pyaudio.PyAudio()
	import warnings
	warnings.filterwarnings("ignore")
	print("Loading app code.")
except:
	print("Import error."); sys.exit()


#==================================================================
# CONFIGURATION

def inputtextlines(filename):
	handle = open(filename,'r')
	linelist = handle.readlines()
	handle.close()
	return linelist

print("Loading configuration file.")
# Assign configuration parameters to variables
try:
	conflines = [
		re.sub(" ","",line.rstrip()).split("=")
		for line in inputtextlines("01-waveform.conf")
		if line[0] != '#'
		]
	confdict = { line[0] : line[1] for line in conflines if line != [] and line[0] != '' }
	channel = confdict['channel']
	volume = float(confdict['volume'])
	soundout = False if confdict['soundout'] == 'False' else True
	figwidth = int(confdict['figwidth'])
	figheight = int(confdict['figheight'])
	fontsize = int(confdict['fontsize'])
	signalcolor = confdict['signalcolor']
	signalabscolor = confdict['signalabscolor']
except:
	print("Configuration error."); sys.exit()

#==============================================================
# FUNCTION DEFINITIONS
#==============================================================
# Sound output

def soundoutput(p, fs, signal, volume):
	product = signal.astype(np.float32)

# for paFloat32 sample values must be in range [-1.0, 1.0]

	stream = p.open(format=pyaudio.paFloat32,
			channels=1,
			rate=fs,
			output=True)
# play
	stream.write((product*volume).tobytes())
	stream.stop_stream()
	stream.close()
	p.terminate()
	return

#==================================================================
# SPEECH INPUT
print("Loading audio file.")
try:
	if len(sys.argv) < 2:
		root = Tk()
		root.withdraw()
		wavfilename = askopenfilename(filetypes=[("Audio files","*.wav")])
		root.update()
		root.destroy()
	else:
		wavfilename = sys.argv[1]

	fs, signal = wave.read(wavfilename)
	sigsecs = 1.0*fs/len(signal)

#==========================
# Process mono-stereo input

	if len(signal.shape) > 1:
		if channel == 'left':
			signal = signal[:,0]
		elif channel == 'right':
			signal = signal[:,1]
		elif channel == 'merge':
			signal = (signal[:,0]/2.0 + signal[:,1]/2.0)
		else:
			print("Unknown signal shape."); sys.exit()
	else:
		channel = 'mono'
except:
	print("Signal input or stereo detection failed."); sys.exit()

# Normalise the signal amplitude to min -1 and max 1; rectify the signal
try:
	signalmax = float(np.max(abs(signal)))
	signal = signal / signalmax
	signalabs = abs(signal)
except:
	print("Normalisation error."); sys.exit()

#==================================================================
# Basic graph display

try:
# Define text strings
	wavfilebase = re.sub('.wav','',wavfilename).split("/")[-1]
	figtitle = "Waveform (%s) [file: "%channel + wavfilebase + "]"
	filetitle = "PNG/" + wavfilebase + "-waveform.png"

# Define the graph
	_,(plt01) = plt.subplots(nrows=1, ncols=1, figsize=(figwidth, figheight))

# Draw the plot
	x = np.arange(len(signal)) / float(fs)
	plt01.plot(x, signal, color=signalcolor, linewidth=1)
	plt01.plot(x, signalabs, color=signalabscolor, linewidth=1)

# Decorate the plot
	plt01.set_title(figtitle, fontsize=fontsize)
	plt01.set_xlabel("Time", fontsize=fontsize)
	plt01.set_ylabel("Amplitude", fontsize=fontsize)
	plt01.set_xlim(0,float(np.ceil(x[-1])))
	plt01.grid(which='both')

# Produce the graph output as file and screen window
	plt.tight_layout(pad=1, w_pad=1, h_pad=1)
	plt.savefig(filetitle)
	plt.show()

except:
	print("Graphics error. Check the configuration."); sys.exit()

#==================================================================
# Sound

if soundout:
	soundoutput(p, fs, signal, volume)

#==============================================================

