#!/usr/bin/python3
# -*- coding: utf-8 -*-
# envelope.py
# D. Gibbon
# 2019-04-20
#==================================================================

try:
	print("Importing sys, re.")
	import sys, re
	ver = sys.version[0]
	if ver == "3":
		print("Using Python3.")
		from tkinter import Tk
		from tkinter.filedialog import askopenfilename
	elif ver == "2":
		print("Using Python2. But Python 3 is required by most ProsodySuite apps.")
		from Tkinter import Tk
		from tkFileDialog import askopenfilename
	else:
		print("Unsupported Python version %s"%ver); exit()
	print("Importing from numpy, matplotlib, scipy.")
	import numpy as np
	import matplotlib.pyplot as plt
	import scipy.io.wavfile as wave
	from scipy.signal import medfilt, butter, lfilter, hilbert
	import pyaudio
	p = pyaudio.PyAudio()
	import warnings
	warnings.filterwarnings("ignore")
	print("Loading app code.")
except:
	print("Import error."); sys.exit()

#==================================================================
#==============================================================
# CONFIGURATION

def inputtextlines(filename):
	handle = open(filename,'r')
	linelist = handle.readlines()
	handle.close()
	return linelist

print("Loading configuration file.")
# Assign configuration parameters to variables
try:
	conflines = [
		re.sub(" ","",line.rstrip()).split("=")
		for line in inputtextlines("02-envelope.conf")
		if line[0] != '#'
		]
	confdict = { line[0] : line[1] for line in conflines if line != [] and line[0] != '' }
	envsuperpose = False if confdict['envsuperpose'] == 'False' else True
	sigsuperpose = False if confdict['sigsuperpose'] == 'False' else True
	channel = confdict['channel']
	volume = float(confdict['volume'])
	soundout = False if confdict['soundout'] == 'False' else True
	envwin = float(confdict['envwin'])
	envmedianfilt = float(confdict['envmedianfilt'])
	envheight = float(confdict['envheight'])
	figwidth = int(confdict['figwidth'])
	figheight = int(confdict['figheight'])
	fontsize = int(confdict['fontsize'])
	signalcolor = confdict['signalcolor']
	signalabscolor = confdict['signalabscolor']
	envelopecolor = confdict['envelopecolor']
except:
	print("Configuration error."); sys.exit()

#==============================================================
# FUNCTION DEFINITIONS
#==============================================================
# Sound output

def soundoutput(p, signal, volume):
	product = signal.astype(np.float32)

# for paFloat32 sample values must be in range [-1.0, 1.0]

	stream = p.open(format=pyaudio.paFloat32,
			channels=1,
			rate=fs,
			output=True)
# play
	stream.write((product*volume).tobytes())
	stream.stop_stream()
	stream.close()
	p.terminate()
	return

#==============================================================

# Signal filter operations: low pass and high pass types
def butterworthfilter(signaldata, cutoff, order, fs, type):
	nyqvist = 0.5 * fs
	normal_cutoff = cutoff / nyqvist
	b, a = butter(order, normal_cutoff, btype=type, analog=False)
	filteredsignal = lfilter(b, a, signaldata)
	return filteredsignal

# Polynomial regression line
def regressionline(x,y,d):
	x = range(len(y))
	fit, res, _, _, _ = np.polyfit(x, y, d, full=True)
	yfit = np.polyval(fit,x)
	return yfit

#==================================================================
# Envelope extraction, Hilbert substitute: get signal peaks (max in window)
# Convert this into a function: positiveenvelope()

def makeenvelope(signal,peakwin,medwin):
	signalabs = abs(signal)
	peaksrange = range(len(signalabs)-peakwin)
	envelope = [ np.max(signalabs[i:i+peakwin]) for i in peaksrange ]
	padleft = [envelope[0]] * int(round(peakwin/2.0))
	padright = [envelope[-1]] * int(round(peakwin/2.0))
	envelope = padleft + envelope + padright
	envelope = medfilt(envelope,medwin)
	envelope = envelope / float(np.max(envelope))
	return envelope

#==============================================================
# SPEECH INPUT
print("Loading audio file.")

try:
	if len(sys.argv) < 2:
		root = Tk()
		root.withdraw()
		wavfilename = askopenfilename(filetypes=[("Audio files","*.wav")])
		root.update()
		root.destroy()
	else:
		wavfilename = sys.argv[1]

	fs, signal = wave.read(wavfilename)
	signallen = len(signal)

except:
	print("Usage: python envelope.py audiofile.wav"); sys.exit()

print("App starting. Wait, processing may take several seconds.")

#==========================
# Process mono-stereo input
try:
	if len(signal.shape) > 1:
		if channel == 'left':
			signal = signal[:,0]
		elif channel == 'right':
			signal = signal[:,1]
		elif channel == 'merge':
			signal = (signal[:,0]/2.0 + signal[:,1]/2.0)
		else:
			print("Unknown signal shape."); sys.exit()
	else:
		channel = 'mono'
except:
	print("Signal input or stereo detection failed."); sys.exit()

# Normalise the signal amplitude to min -1 and max 1
try:
	signalmax = float(np.max(abs(signal)))
	signal = signal / signalmax
except:
	print("Normalisation error."); sys.exit()

#==============================================================
# Envelope extraction

try:
	sigabs = np.abs(signal)

	envsamp = int(fs * envwin)
	envmediansamp = int(fs * envmedianfilt)
	envmediansamp = envmediansamp + envmediansamp % 2 - 1 # ensure odd
	envelope = makeenvelope(signal,envsamp,envmediansamp)
	envelope = envheight * envelope / np.max(envelope)
#	envelope = butterworthfilter(envelope, 50, 11, fs, "low")

except:
	print("Error in envelope transformation"); sys.exit()

#==============================================================
# GRAPHICS

try:

# Define text strings
	wavfilebase = re.sub('.wav','',wavfilename).split("/")[-1]
	figtitle = "Envelope extraction [file: " + wavfilebase + "]"
	filetitle = "PNG/" + wavfilebase + "-envelope.png"

# Create the graph format
	_,(plt01, plt02) = plt.subplots(nrows=2,ncols=1,figsize=(figwidth,figheight))
	plt.suptitle(figtitle, fontsize=fontsize)

# Draw the top plot (waveform)
	x = np.arange(len(signal)) / float(fs)
	plt01.plot(x, signal, color=signalcolor)
	if envsuperpose: plt01.plot(x,envelope, linewidth = 2, color=envelopecolor)
	plt01.set_title("Waveform oscillogram (%s)"%channel, fontsize=fontsize)
	plt01.set_xlabel("Time (s)", fontsize=fontsize)
	plt01.set_ylabel("Amplitude", fontsize=fontsize)
	plt01.set_xlim(0,float(np.ceil(x[-1])))
	plt01.grid()

# Draw the lower plot (amplitude envelope)
	plt02.set_title("Rectified (absolute) signal and Amplitude Envelope (Peak-Picker: peakwin=%.4f, medfiltwin=%d)"%(envwin, envmedianfilt), fontsize=fontsize)
	plt02.set_xlabel("Time", fontsize=fontsize)
	plt02.set_ylabel("Amplitude", fontsize=fontsize)
	if sigsuperpose: plt02.plot(x, sigabs, color=signalabscolor)
	x = np.arange(0,len(envelope)) / (1.0 * fs)
	plt02.plot(x, envelope, linewidth=2, color=envelopecolor)
	plt02.set_ylim(0, 1.05)
	plt02.set_xlim(0,float(np.ceil(x[-1])))
	plt02.grid()

# Configure and produce the graph output as file and screen window
	plt.tight_layout(pad=3, w_pad=1, h_pad=1)
	plt.savefig(filetitle)
	plt.show()

except:
	print("Graphics error. Check the configuration."); sys.exit()

#==============================================================

if soundout:
	soundoutput(p, signal, volume)
