#!/usr/bin/python3
# -*- coding: utf-8 -*-
# f0.py
# D. Gibbon
# 2019-04-20
#==================================================================

try:
	print("Importing sys, re.")
	import sys, re
	ver = sys.version[0]
	if ver == "3":
		print("Using Python3.")
		from tkinter import Tk
		from tkinter.filedialog import askopenfilename
	elif ver == "2":
		print("Using Python2. But Python 3 is required by most ProsodySuite apps.")
		from Tkinter import Tk
		from tkFileDialog import askopenfilename
	else:
		print("Unsupported Python version %s"%ver); exit()
	print("Importing from numpy, matplotlib, scipy.")
	import numpy as np
	import matplotlib.pyplot as plt
	import scipy.io.wavfile as wave
	from scipy.signal import medfilt, butter, lfilter, hilbert
	import pyaudio
	p = pyaudio.PyAudio()
	import warnings
	warnings.filterwarnings("ignore")
	print("Loading app code.")
except:
	print("Import error."); sys.exit()


#==================================================================
#==============================================================
# CONFIGURATION

def inputtextlines(filename):
	handle = open(filename,'r')
	linelist = handle.readlines()
	handle.close()
	return linelist

print("Loading configuration file.")
# Assign configuration parameters to variables
if True:
	conflines = [
		re.sub(" ","",line.rstrip()).split("=")
		for line in inputtextlines("03-f0.conf")
		if line[0] != '#'
		]
	confdict = { line[0] : line[1] for line in conflines if line != [] and line[0] != '' }
	channel = confdict['channel']
	volume = float(confdict['volume'])
	soundout = False if confdict['soundout'] == 'False' else True
	f0min = int(confdict['f0min'])
	f0max = int(confdict['f0max'])
	highpass = int(confdict['highpass'])
	highpassorder = int(confdict['highpassorder'])
	lowpass = int(confdict['lowpass'])
	lowpassorder = int(confdict['lowpassorder'])
	framerate = float(confdict['framerate'])
	winoffsetdivisor = int(confdict['winoffsetdivisor'])
	f0medianfilt = int(confdict['f0medianfilt'])
	f0globalpoly = int(confdict['f0globalpoly'])
	figwidth = int(confdict['figwidth'])
	figheight = int(confdict['figheight'])
	signalcolor = confdict['signalcolor']
	f0color = confdict['f0color']
	f0polymodelcolor = confdict['f0polymodelcolor']
	fontsize = int(confdict['fontsize'])
if False:
	print("Configuration error."); sys.exit()

#==============================================================
# FUNCTION DEFINITIONS
#==============================================================
# Sound output

def soundoutput(p, fs, signal, volume):
	product = signal.astype(np.float32)

# for paFloat32 sample values must be in range [-1.0, 1.0]

	stream = p.open(format=pyaudio.paFloat32,
			channels=1,
			rate=fs,
			output=True)
# play
	stream.write((product*volume).tobytes())
	stream.stop_stream()
	stream.close()
	p.terminate()
	return

#==============================================================
# Fourier transformation

def fft(signal,fs):	# signal is a NumPy array; fs is sampling rate integer.
	period = 1.0/fs
	mags = np.abs(np.fft.rfft(signal))**2
	freqs = np.abs(np.fft.rfftfreq(signal.size,period))
	mags = np.asarray([0.000001 if m==0 else m for m in mags])
# Flatten magnitudes, remove initial zero
	maglog = np.log10(mags)
	maglog[0] = maglog[1]
# Amplify and normalise magnitudes
	maglog = maglog**2
	maglog = maglog / np.max(maglog)
	return np.asarray(freqs), np.asarray(maglog)

# Signal filter operations: low pass and high pass types
def butterworthfilter(signaldata, cutoff, order, fs, type):
	nyqvist = 0.5 * fs
	normal_cutoff = cutoff / nyqvist
	b, a = butter(order, normal_cutoff, btype=type, analog=False)
	filteredsignal = lfilter(b, a, signaldata)
	return filteredsignal

# Polynomial regression line
def regressionline(x,y,d):
	x = range(len(y))
	fit, res, _, _, _ = np.polyfit(x, y, d, full=True)
	yfit = np.polyval(fit,x)
	return yfit

# Average Magnitude Difference Function
def amdf(fs, signal, signallen, framerate, framelen, diffoffset, f0medianfilt):
	rangelen = signallen-framelen
	irange = range(0, rangelen, framelen)
	f0listlen = int(float(rangelen)/framelen)
	f0list = [0]*(1+f0listlen)
	for i, framestart in enumerate(irange):
		framestop = framestart + framelen
		meandiffs = [0]*framelen
		for j, winstart in enumerate(range(framestart,framestop)):
			winstop = winstart+framelen
			frame = signal[framestart:framestop]
			win = signal[winstart:winstop]
			if len(frame) != len(win):
				break
			absdiffs = abs(frame - win)
			meandiffs[j] = np.mean(absdiffs)
		smallest = min(meandiffs[diffoffset:])
		index = meandiffs.index(smallest)
		if index > 0: f0 = fs / index
		else: f0 = 0
		f0list[i] = 0 if (f0 < f0min or f0 > f0max) else f0
	f0list = medfilt(f0list, f0medianfilt)
	return f0list

#==============================================================
# SPEECH INPUT
print("Loading audio file.")
try:
	if len(sys.argv) < 2:
		root = Tk()
		root.withdraw()
		wavfilename = askopenfilename(filetypes=[("Audio files","*.wav")])
		root.update()
		root.destroy()
	else:
		wavfilename = sys.argv[1]
	fs, signal = wave.read(wavfilename)
	signallen = len(signal)
	signalsecs = float(signallen)/fs

#==========================
# Process mono-stereo input

	if len(signal.shape) > 1:
		if channel == 'left':
			signal = signal[:,0]
		elif channel == 'right':
			signal = signal[:,1]
		elif channel == 'merge':
			signal = (signal[:,0]/2.0 + signal[:,1]/2.0)
		else:
			print("Unknown signal shape."); sys.exit()
	else:
		channel = 'mono'
except:
	print("Signal input or stereo detection failed."); sys.exit()

# Normalise the signal amplitude to min -1 and max 1
try:
	signalmax = float(np.max(abs(signal)))
	signal = signal / signalmax
except:
	print("Normalisation error."); sys.exit()

# F0 estimation ('pitch tracking')

try:

	framelen = int(round(fs * framerate))
	framecount = int(round(1.0 * signallen / framelen))
	diffoffset = int(round(framelen/winoffsetdivisor))

	# F0 preprocessing
	signalfiltered = butterworthfilter(signal, highpass, highpassorder, fs, "high")
	signalfiltered = butterworthfilter(signalfiltered, lowpass, lowpassorder, fs, "low")

	# F0 estimation
	f0list = amdf(fs, signalfiltered, signallen, framerate, framelen, diffoffset, f0medianfilt)

	f0nozeros = np.asarray( [ f for f in f0list if f > 0 ] )

	# F0 postprocessing
	f0median = np.median(f0nozeros)

	f0medianised = f0list.copy()
	for i,f in enumerate(f0medianised):
		if f==0: f0medianised[i] = f0median

except:
	print("Error in F0 estimation"); sys.exit()

#==============================================================
# GRAPHICS

try:

# Define text strings
	wavfilebase = re.sub('.wav','',wavfilename).split("/")[-1]
	figtitle = "F0 estimation [file: " + wavfilebase + "]"
	filetitle = "PNG/" + wavfilebase + "-f0.png"

# Create the graph format
	_,(plt01, plt02) = plt.subplots(nrows=2,ncols=1,figsize=(figwidth,figheight))
	plt.suptitle(figtitle)
# Draw the top plot (waveform)
	x1 = np.arange(len(signal)) / float(fs)
	plt01.plot(x1, signal, color=signalcolor)
	plt01.set_title("Waveform oscillogram (%s)"%channel)
	plt01.set_xlabel("Time (s)")
	plt01.set_ylabel("Amplitude")
	plt01.set_xlim(0,float(np.ceil(signalsecs)))
	plt01.grid()

# Draw the lower plot (long term spectrum)
	plt02.set_title("AMDF Fundamental frequency estimation (F0, 'pitch tracking')", fontsize=fontsize)
	plt02.set_xlabel("Time (s)", fontsize=fontsize)
	plt02.set_ylabel("Frequency (Hz)", fontsize=fontsize)
	plt02.grid()
	x2 = np.arange(0,len(f0list)) * framerate
	f0polymodel = regressionline(x2,f0medianised,f0globalpoly)
	plt02.plot(x2,f0polymodel, color=f0polymodelcolor)
	plt02.scatter(x2,f0list, s=4, color=f0color)
	plt02.set_xlim(0,float(np.ceil(signalsecs)))
	plt02.set_ylim(f0min, f0max)

# Configure and produce the graph output as file and screen window
	plt.tight_layout(pad=3, w_pad=1, h_pad=1)
	plt.savefig(filetitle)
	plt.show()

except:
	print("Graphics error. Check the configuration."); sys.exit()

#==============================================================
# Sound

if soundout:
	soundoutput(p, fs, signal, volume)

#==============================================================

