#!/usr/bin/python3
# -*- coding: utf-8 -*-
# ltas.py
# D. Gibbon
# 2019-04-20
#==================================================================

try:
	print("Importing sys, re.")
	import sys, re
	ver = sys.version[0]
	if ver == "3":
		print("Using Python3.")
		from tkinter import Tk
		from tkinter.filedialog import askopenfilename
	elif ver == "2":
		print("Using Python2. But Python 3 is required by most ProsodySuite apps.")
		from Tkinter import Tk
		from tkFileDialog import askopenfilename
	else:
		print("Unsupported Python version %s"%ver); exit()
	print("Importing from numpy, matplotlib, scipy.")
	import numpy as np
	import matplotlib.pyplot as plt
	import scipy.io.wavfile as wave
	from scipy.signal import medfilt, butter, lfilter, hilbert
	import pyaudio
	p = pyaudio.PyAudio()
	import warnings
	warnings.filterwarnings("ignore")
	print("Loading app code.")
except:
	print("Import error."); sys.exit()
#==============================================================
# CONFIGURATION

def inputtextlines(filename):
	handle = open(filename,'r')
	linelist = handle.readlines()
	handle.close()
	return linelist

print("Loading configuration file.")
# Assign configuration parameters to variables
try:
	conflines = [
		re.sub(" ","",line.rstrip()).split("=")
		for line in inputtextlines("05-ltas.conf")
		if line[0] != '#'
		]
	confdict = { line[0] : line[1] for line in conflines if line != [] and line[0] != '' }
	channel = confdict['channel']
	volume = float(confdict['volume'])
	soundout = False if confdict['soundout'] == 'False' else True
	figwidth = int(confdict['figwidth'])
	figheight = int(confdict['figheight'])
	fontsize = int(confdict['fontsize'])
	signalcolor = confdict['signalcolor']
	ltasfreqsplotcolor = confdict['ltasfreqsplotcolor']
	ltasfreqsscattercolor = confdict['ltasfreqsscattercolor']
	ltasminfreq = int(confdict['ltasminfreq'])
	ltasmaxfreq = int(confdict['ltasmaxfreq'])
	ltasmedfilt = int(confdict['ltasmedfilt'])
except:
	print("Configuration error."); sys.exit()

#==============================================================
# FUNCTION DEFINITION: Fourier transformation
#==============================================================
# Sound output

def soundoutput(p, signal, volume):
	product = signal.astype(np.float32)

# for paFloat32 sample values must be in range [-1.0, 1.0]

	stream = p.open(format=pyaudio.paFloat32,
			channels=1,
			rate=fs,
			output=True)
# play
	stream.write((product*volume).tobytes())
	stream.stop_stream()
	stream.close()
	p.terminate()
	return

#==============================================================

def fft(signal,fs):	# signal is a NumPy array; fs is sampling rate integer.

	period = 1.0/fs
	mags = np.abs(np.fft.rfft(signal))**2
	freqs = np.abs(np.fft.rfftfreq(signal.size,period))
	mags = np.asarray([0.000001 if m==0 else m for m in mags])

# Flatten magnitudes, remove initial zero
	maglog = np.log10(mags)
	maglog[0] = maglog[1]

# Amplify magnitudes
	maglog = maglog**2
	maglog = maglog / np.max(maglog)

	return freqs, maglog

#==============================================================
# SPEECH INPUT
print("Loading audio file.")
try:
	if len(sys.argv) < 2:
		root = Tk()
		root.withdraw()
		wavfilename = askopenfilename(filetypes=[("Audio files","*.wav")])
		root.update()
		root.destroy()
	else:
		wavfilename = sys.argv[1]

	fs, signal = wave.read(wavfilename)

except:
	print("Usage: python ltas.py audiofile.wav"); sys.exit()

print("App starting. Wait, processing may take several seconds.")

#==========================
# Process mono-stereo input

try:
	if len(signal.shape) > 1:

		if channel == 'left':
			signal = signal[:,0]
		if channel == 'right':
			signal = signal[:,1]
		if channel == 'merge':
			signal = (signal[:,0]/2.0 + signal[:,1]/2.0)
	else:
		channel = 'mono'
except:
	print("Signal input or stereo detection failed."); sys.exit()

# Normalise the signal amplitude to min -1 and max 1, calculate long term spectrum
try:
	signalmax = float(np.max(abs(signal)))
	signal = signal / signalmax

	ltasfreqs,ltasmags = fft(abs(signal),fs)
	ltasmags = medfilt(ltasmags,ltasmedfilt)**3
	ltasmagdiffs = np.diff(ltasmags)**3

	ltasmags = ltasmags/np.max(ltasmags)
	ltasmagdiffs = ltasmagdiffs/np.max(ltasmagdiffs)
except:
	print("Error in signal processing."); sys.exit()

#==============================================================
# GRAPHICS

try:

# Define text strings
	wavfilebase = re.sub('.wav','',wavfilename).split("/")[-1]
	figtitle = "Long Term Amplitude Envelope Spectrum [file: " + wavfilebase + "]"
	filetitle = "PNG/" + wavfilebase + "-ltas.png"

# Create the graph format
	_,(plt01, plt02) = plt.subplots(nrows=1,ncols=2,figsize=(figwidth,figheight))

	plt.suptitle(figtitle, fontsize=fontsize)

# Draw the top plot (waveform)
	x = np.arange(len(signal)) / float(fs)
	plt01.plot(x, signal, color=signalcolor)
	plt01.set_title("Waveform oscillogram (%s)"%channel, fontsize=fontsize)
	plt01.set_xlabel("Time (s)", fontsize=fontsize)
	plt01.set_ylabel("Amplitude", fontsize=fontsize)
	plt01.set_xlim(0,float(np.ceil(x[-1])))
	plt01.grid()

# Draw the lower plot (long term spectrum)
	plt02.set_title("Long Term Amplitude Spectrum", fontsize=fontsize)
	plt02.plot(ltasfreqs, ltasmags, color=ltasfreqsplotcolor)
	plt02.scatter(ltasfreqs, ltasmags, color=ltasfreqsscattercolor)
	plt02.set_xlim(ltasminfreq, ltasmaxfreq)
	plt02.set_xlabel("Frequency (Hz)", fontsize=fontsize)
	plt02.set_ylabel("Magnitude", fontsize=fontsize)
	plt02.grid()

# Configure and produce the graph output as file and screen window
	plt.tight_layout(pad=3, w_pad=1, h_pad=1)
	plt.savefig(filetitle)
	plt.show()

except:
	print("Graphics error. Check the configuration."); sys.exit()

#==============================================================
# Sound

if soundout:
	soundoutput(p, signal, volume)

#==============================================================


