#!/usr/bin/python3
# -*- coding: utf-8 -*-
# prosody.py
# D. Gibbon
# 2019-04-10
# Input: wav speech file
# Output: PNG graph with speech signal visualisations
# Visualisations:
#	- Amplitude waveform
#	- Amplitude envelope
#	- Fundamental frequency (F0)
#	- Spectrogram
#	- Long term spectrum
#	- Long term envelope spectrum
#	- Long term F0 spectrum
#================================================================

try:
	print("Importing sys, re.")
	import sys, re
	ver = sys.version[0]
	if ver == "3":
		print("Using Python3.")
		from tkinter import Tk
		from tkinter.filedialog import askopenfilename
	elif ver == "2":
		print("Using Python2. But Python 3 is required by most ProsodySuite apps.")
		from Tkinter import Tk
		from tkFileDialog import askopenfilename
	else:
		print("Unsupported Python version %s"%ver); exit()
	print("Importing from numpy, matplotlib, scipy.")
	import numpy as np
	import matplotlib.pyplot as plt
	import scipy.io.wavfile as wave
	from scipy.signal import medfilt, butter, lfilter, hilbert
	import pyaudio
	p = pyaudio.PyAudio()
	import warnings
	warnings.filterwarnings("ignore")
	print("Loading app code.")
except:
	print("Import error."); sys.exit()

#==============================================================
# CONFIGURATION

def inputtextlines(filename):
	handle = open(filename,'r')
	linelist = handle.readlines()
	handle.close()
	return linelist

print("Loading configuration file.")
# Assign configuration parameters to variables
try:
	conflines = [
		re.sub(" ","",line.rstrip()).split("=")
		for line in inputtextlines("07-prosody.conf")
		if line[0] != '#'
		]
	confdict = { line[0] : line[1] for line in conflines if line != [] and line[0] != '' }
	channel = confdict['channel']
	volume = float(confdict['volume'])
	soundout = False if confdict['soundout'] == 'False' else True

	graphwidth = int(confdict['graphwidth'])
	graphheight = int(confdict['graphheight'])
	fontsize = int(confdict['fontsize'])

	signalcolor = confdict['signalcolor']
	abssignalcolor = confdict['abssignalcolor']
	envelopecolor = confdict['envelopecolor']
	f0color = confdict['f0color']
	f0polymodelcolor = confdict['f0polymodelcolor']
	medianlinecolor = confdict['medianlinecolor']

	ltasfreqsplotcolor = confdict['ltasfreqsplotcolor']
	ltasfreqsscattercolor = confdict['ltasfreqsscattercolor']
	envfreqsplotcolor = confdict['envfreqsplotcolor']
	envfreqsscattercolor = confdict['envfreqsscattercolor']
	ltfsfreqsplotcolor = confdict['ltfsfreqsplotcolor']
	ltfsfreqsscattercolor = confdict['ltfsfreqsscattercolor']

	waveformshow = False if confdict['waveformshow'] == 'False' else True

	f0min = int(confdict['f0min'])
	f0max = int(confdict['f0max'])

	spectrogramfmin = int(confdict['spectrogramfmin'])
	spectrogramfmax = int(confdict['spectrogramfmax'])
	spectrogramres = int(confdict['spectrogramres'])

	f0globalpoly = int(confdict['f0globalpoly'])
	spectrumpoly = float(confdict['spectrumpoly'])

	rhythmcount = int(confdict['rhythmcount'])
	rhythmlabel = False if confdict['rhythmlabel'] == 'False' else True
	medianline = False if confdict['medianline'] == 'False' else True

	ltsmin = int(confdict['ltsmin'])
	ltsmax = int(confdict['ltsmax'])
	ltsmedfilt = int(confdict['ltsmedfilt'])

	aemsmin = int(confdict['aemsmin'])
	aemsmax = int(confdict['aemsmax'])
	aemsmedfilt = int(confdict['aemsmedfilt'])
	femsmin = float(confdict['femsmin'])
	femsmax = int(confdict['femsmax'])
	femsmedfilt = int(confdict['femsmedfilt'])

	envwin = int(confdict['envwin'])
	envmedianfilt = int(confdict['envmedianfilt'])

	centrethresh = float(confdict['centrethresh'])
	limitthresh = float(confdict['limitthresh'])
	framerate = float(confdict['framerate'])
	winoffsetdivisor = int(confdict['winoffsetdivisor'])
	f0medianfilt = int(confdict['femsmedfilt'])

	infotext = u'#1 Waveform → #5 Long Term Spectrum\n\n#1 Waveform → #2 Envelope → #6 Envelope spectrum\n\n#1 Waveform → #3 F0 → #7 F0 spectrum'

except:
	print("Configuration error."); sys.exit()

#================================================================
# Type check of some variables

if np.mod(ltsmedfilt,2)==0 or np.mod(aemsmedfilt,2)==0 or np.mod(femsmedfilt,2)==0:
	print("Median filter error: assign odd number."); sys.exit()
#================================================================
# FUNCTION DEFINITIONS
#================================================================
# Sound output

def soundoutput(p, signal, volume):
	product = signal.astype(np.float32)

# for paFloat32 sample values must be in range [-1.0, 1.0]

	stream = p.open(format=pyaudio.paFloat32,
			channels=1,
			rate=fs,
			output=True)
# play
	stream.write((product*volume).tobytes())
	stream.stop_stream()
	stream.close()
	p.terminate()
	return

#==============================================================

# Low pass and high pass filters

# Call example:
# butterworthfilter(signaldata,80, 3, 16000, 'low')
# butterworthfilter(signaldata,2000, 3, 16000, 'low')

def butterworthfilter(signaldata, cutoff, order, fs, type):
	nyqvist = 0.5 * fs
	normal_cutoff = cutoff / nyqvist
	b, a = butter(order, normal_cutoff, btype=type, analog=False)
	filteredsignal = lfilter(b, a, signaldata)
	return filteredsignal


#================================================================

# Signal clipping

def clipper(sig,thresh,type):
	if type == "centre":
		clipped = (abs(sig) > thresh).astype(int) * sig
	elif type == "limit":
		clipped = (abs(sig) < thresh).astype(int) * sig
	elif type == "lower":
		clipped = (sig > thresh).astype(int) * sig
	elif type == "upper":
		clipped = (sig < thresh).astype(int) * sig
	else:
		print("Unknown type:",type); sys.exit()
	return np.asarray(clipped)


#==================================================================
# TIME DOMAIN MODELS
#==================================================================

# Fourier transformation

def fft(signal,fs):	# signal is a NumPy array; fs is sampling rate integer.
	period = 1.0/fs
	mags = np.abs(np.fft.rfft(signal))**2
	freqs = np.abs(np.fft.rfftfreq(signal.size,period))
	mags = np.asarray([0.000001 if m==0 else m for m in mags])
	maglog = np.log10(mags)	# CHECK IF THIS IS NECESSARY
	maglog[0] = maglog[1]
	maglog = maglog**2
	maglog = maglog / np.max(maglog)
	return freqs, maglog

#==================================================================
# Regression line

def polyregline(x,y,d):
	x = range(len(y))
	fit, res, _, _, _ = np.polyfit(x, y, d, full=True)
	yfit = np.polyval(fit,x)
	return yfit
# fit with np.polyfit; linear (degree 1) regression only

# Not used
def polyregline2(x,y,d):
	m, b, _, _, _ = stats.linregress(x, y)
	poly = [ m * xx + b for xx in x]
	return poly

#==================================================================

# Hilbert substitute: get signal peaks (max in window)
# Convert this into a function: positiveenvelope()

def makeenvelope(signal,peakwin,medwin):
	signalabs = abs(signal)
	peaksrange = range(len(signalabs)-peakwin)
	envelope = [ np.max(signalabs[i:i+peakwin]) for i in peaksrange ]
	padleft = [envelope[0]] * int(round(peakwin/2.0))
	padright = [envelope[-1]] * int(round(peakwin/2.0))
	envelope = padleft + envelope + padright
	envelope = medfilt(envelope,medwin)
	envelope = envelope / float(np.max(envelope))
	return envelope

#==================================================================
# Average Magnitude Difference Function

def amdf(fs, signal, signallen, framerate, framelen, diffoffset, f0medianfilt):
	rangelen = signallen-framelen
	irange = range(0, rangelen, framelen)
	f0listlen = int(float(rangelen)/framelen)
	f0list = np.zeros(1+f0listlen)
	for i, framestart in enumerate(irange):
		framestop = framestart + framelen
		meandiffs = np.zeros(framelen).tolist()
		for j, winstart in enumerate(range(framestart,framestop)):
			winstop = winstart+framelen
			frame = signal[framestart:framestop]
			win = signal[winstart:winstop]
			if len(frame) != len(win):
				break
			absdiffs = abs(frame - win)
			meandiffs[j] = np.mean(absdiffs)
		smallest = min(meandiffs[diffoffset:])
		index = meandiffs.index(smallest)
		if index > 0:
			f0 = fs / index
		else:
			f0 = 0
		f0list[i] = 0 if (f0 < f0min or f0 > f0max) else f0
	f0list = medfilt(f0list, f0medianfilt)
	return f0list

#==================================================================
# Time domain graph creation #==================================================================

# Make time domain graph

def timedomaingraph(panel, draw, title, label, samprate, data, ymin, ymax, colour, width):

	lendata = len(data)			# tickcount = lendata
	datasecs = lendata * 1.0/samprate
	x = np.linspace(0,lendata,lendata) * 1.0/float(samprate)
	if draw == "plot":
		panel.plot(x,data, label=label, color=colour, linewidth=width)
	if draw == "scatter":
		panel.scatter(x,data, label=label, color=colour, s=3)
	xtix = panel.get_xticks()
	xtixlabels = [ "%.1fs"%(i+1) for i in xtix ]
	panel.set_xticklabels(xtixlabels,fontsize=fontsize-1)
	panel.set_ylim(ymin, ymax)
	panel.set_title(title, fontsize=fontsize, fontweight='bold')
	panel.set_xlim(0,float(np.ceil(datasecs)))
#	panel.legend(fontsize=fontsize-3)	# Not yet adapted.
	return

#==================================================================
# Frequency domain graph creation
#==================================================================

def frequencydomaingraph(panel, draw, title, label, samprate, xax, data, ymin, ymax, xmin, xmax, colour, width):

	numin = int(round(xmin * len(xax) / xax[-1]))
	numax = int(xmax * len(xax) / xax[-1])
	xax = xax[numin:numax]
	data = data[numin:numax]

	x = np.arange(len(data))
	polymodel = polyregline(x,data,spectrumpoly)
	polyresid = data - polymodel

	if rhythmcount > 0:
		panel.plot(xax,polyresid+abs(np.min(polyresid)), color='grey', label='norm')

	if spectrumpoly > 0:
		panel.plot(xax,polymodel, linewidth=2, color='r', label='polymodel')

	if medianline:
		datamedian = np.median(data)
		panel.axhline(datamedian, linewidth=2, color='pink', label='spectrum median')

	panel.scatter(xax,data, label=label, s=8)

# Rhythm bars
	if rhythmcount > 0:
		a = sorted(polyresid.tolist())
		aa = reversed(a[-rhythmcount:])
		textpos = 0.8
		for item in aa:
			b = polyresid.tolist().index(item)
			f = float(b)/signalsecs
			if f > 0:
				s = 1.0/f
			else:
				s = 0
			c = b/signalsecs
			panel.axvline(f, linewidth=2, color='r')
			if rhythmlabel:
				panel.text(c, textpos, " ".join(["%.1fHz"%f,"%.3fms"%s]),fontsize=fontsize-1, color="r")
				textpos -= 0.2

	panel.set_title(title, fontsize=fontsize, fontweight='bold')
	panel.set_xlim(xmin, xmax)
	panel.set_ylim(ymin-0.1, ymax+0.1)
	xtix = panel.get_xticks()
	xtixlabels = [ "%.1f Hz\n%.3f s"%(i,1.0/i) if i>0 else "0 Hz" for i in xtix ]
	panel.set_xticklabels(xtixlabels,fontsize=fontsize-1)
	panel.legend(fontsize=fontsize-4)
	return

#==================================================================
# MAIN CALLS
#==================================================================
# SPEECH INPUT
print("Loading audio file.")
try:
	if len(sys.argv) < 2:
		root = Tk()
		root.withdraw()
		wavfilename = askopenfilename(filetypes=[("Audio files","*.wav")])
		root.update()
		root.destroy()
	else:
		wavfilename = sys.argv[1]

	wavfilebase = re.sub('.wav','',wavfilename).split("/")[-1]
	figtitle = "Speech Modulations and Models V04 2019-04-18 DG [file: " + wavfilebase + "]"
	filetitle = "PNG/" + wavfilebase + "-prosody.png"
except:
	print("Usage: prosody audiofile.wav"); sys.exit()

print("App starting. Wait, processing may take several seconds.")

#==================================================================
# Signal input (mono file)

try:
	fs, signal = wave.read(wavfilename)	# fs is sampling rate
except:
	print("Input failure."); sys.exit()

#==========================
# Process mono-stereo input

try:
	if len(signal.shape) > 1:
		if channel == 'left':
			signal = signal[:,0]
		elif channel == 'right':
			signal = signal[:,1]
		elif channel == 'merge':
			signal = (signal[:,0]/2.0 + signal[:,1]/2.0)
		else:
			print("Unknown channel selection."); sys.exit()
	else:
		channel = 'mono'

except:
	print("Signal input or stereo detection failed."); sys.exit()

try:
	signal = signal/float(max(signal))
	sampleperiod = 1.0/float(fs)
	signallen = len(signal)
	signalsecs = len(signal) / float(fs)
	wavstart = 0
	wavlength = len(signal)
	wavend = wavstart + wavlength
except:
	print("Signal properties calculation error."); sys.exit()

#==================================================================
# TIME DOMAIN TRANSFORMATIONS
#==================================================================
# Envelope

try:
	envelope = makeenvelope(signal,envwin,envmedianfilt)
except:
	print("Envelope calculation error."); sys.exit()

#==================================================================
# F0 estimation (AMDF algorithm with signal filtering)

try:
	framelen = int(round(fs * framerate))
	framecount = int(round(1.0 * signallen / framelen))
	diffoffset = int(round(framelen/winoffsetdivisor))

	# F0 preprocessing
	clipsig = signal.copy()
	clipsig = clipper(clipsig,centrethresh,"centre")
	clipsig = clipper(clipsig,limitthresh,"limit")
	clipsig = butterworthfilter(clipsig, f0min+15, 5, fs, "high")
	clipsig = butterworthfilter(clipsig, f0max-10, 2, fs, "low")

	# F0 estimation
	f0list = amdf(fs, clipsig, wavlength, framerate, framelen, diffoffset, f0medianfilt)

	f0nozeros = np.asarray( [ f for f in f0list if f > f0min ] )
	f0zeroised = np.asarray( [ f if f > f0min else 0 for f in f0list ] )

	# F0 postprocessing
	f0median = np.median(f0nozeros)
	f0medianised = np.asarray( [ f if f > f0min else f0median for f in f0list ] )

except:
	print("F0 calculation error."); sys.exit()

#==================================================================
# FREQUENCY DOMAIN TRANSFORMATIONS
#==================================================================

# Long term spectrum

try:
	ltsf0,ltsmags = fft(abs(signal),fs)
	ltsmags = medfilt(ltsmags,ltsmedfilt)
	ltsmagsdiff = np.diff(ltsmags)
	ltsmagsdiff = ltsmagsdiff/float(np.max(ltsmagsdiff))
except:
	print("FEMS calculation error."); sys.exit()

#==================================================================
# Amplitude Envelope Modulation Spectrum

try:
	aemsf0,aemsmags = fft(abs(envelope),fs)
	aemsmags = medfilt(aemsmags,aemsmedfilt)
	aemsmagsdiff = np.diff(aemsmags)

	aemsmagsdiff = aemsmagsdiff/float(np.max(aemsmagsdiff))
except:
	print("AEMS calculation error."); sys.exit()

#==================================================================
# Fundamental Frequency Envelope Modulation Spectrum

try:
	femsf0,femsmags = fft(f0list,1.0/framerate)
	femsmags = medfilt(femsmags,femsmedfilt)
	femsmagsdiff = np.diff(femsmags)

	femsmags = femsmags/float(np.max(femsmags))
	femsmagsdiff = femsmagsdiff/float(np.max(femsmagsdiff))
except:
	print("F0 spectrum calculation error."); sys.exit()

#==================================================================
#==================================================================
# INITIALISE GRAPHS
#==================================================================

try:
	_,(	(pltwave, pltltspec),	(pltenv, pltaems),	(pltf0, pltfems), (pltspectro, plttext)
	) = plt.subplots(nrows=4,ncols=2,figsize=(graphwidth,graphheight))

	plt.suptitle(figtitle, fontsize=fontsize+2, fontweight='bold')
except:
	print("Graph initialization error."); sys.exit()

#==================================================================
# COLUMN 1 - Time domain graphs
#==================================================================
# ROW 1 - Waveform (oscillogram)

try:
	timedomaingraph(pltwave, "plot", "1. Waveform (%s)"%channel, "waveform", fs, signal, -1.0, 1.0, signalcolor, 0.5)
	pltwave.set_ylabel("Amplitude", fontsize=fontsize-1)
except:
	print("Oscillogram graph error."); sys.exit()

#==================================================================
# ROW 2 - Amplitude envelope

try:
	if waveformshow:
		timedomaingraph(pltenv, "plot", "2. Amplitude Envelope (Emulation of Absolute Hilbert Transform)", "waveform", fs, signal, -1.0, 1.0, signalcolor, 0.5)
		timedomaingraph(pltenv, "plot", "2. Amplitude Envelope (Emulation of Absolute Hilbert Transform)", "waveform", fs, abs(signal), -1.0, 1.0, abssignalcolor, 0.5)
	
	timedomaingraph(pltenv, "plot", "2. Amplitude Envelope (Emulation of Absolute Hilbert Transform)", "envelope", fs, envelope, -1.0, 1.0, envelopecolor,2)
	pltenv.set_ylabel("Amplitude)", fontsize=fontsize-1)
except:
	print("Amplitude envelope graph error."); sys.exit()

#==================================================================
# ROW 3 - Fundamental frequency estimation

try:

	timedomaingraph(pltf0, "scatter", "3. Fundamental Frequency (F0, 'pitch') estimate, AMDF algorithm", "F0", 1.0/framerate, f0zeroised, f0min, f0max, f0color, 0.1)

	if medianline:
		pltf0.axhline(f0median, linewidth=2, color=medianlinecolor, label='median')

	if f0globalpoly > 0:
		x = np.arange(len(f0medianised))*1.0/framerate
		polymodel = polyregline(x,f0medianised, f0globalpoly)
		timedomaingraph(pltf0, "scatter", "3. Fundamental Frequency (F0, 'pitch') estimate, AMDF algorithm", "F0", 1.0/framerate, polymodel, f0min, f0max, f0polymodelcolor, 0.1)
	
	pltf0.set_ylabel("Frequency (Hz)", fontsize=fontsize-1)

except:
	print("F0 graph error."); sys.exit()

#==================================================================
# ROW 4 - Spectrogram

try:
	pltspectro.set_title("4. Spectrogram (sequence of short-term Spectra (FFTs))", fontsize=fontsize, fontweight='bold')
	x = np.arange(len(signal)) / float(fs)
	pltspectro.set_title("Spectrogram (spectrum sequence)")
	pltspectro.specgram(signal, NFFT=spectrogramres, Fs=fs,cmap='jet')
	pltspectro.set_ylim(spectrogramfmin,spectrogramfmax)
	pltspectro.set_ylabel("Frequency (Hz)")
	pltspectro.set_xlim(0,float(np.ceil(x[-1])))
	xtix = pltspectro.get_xticks()
	xtixlabels = [ "%.1f s"%(i) if i>0 else "0 s" for i in xtix ]
	pltspectro.set_xticklabels(xtixlabels,fontsize=fontsize-1)
	pltspectro.grid(which='both')

except:
	print("Spectrogram error."); sys.exit()

#==================================================================
# COLUMN 2 - Frequency domain graphs
#==================================================================
# ROW 1 - Long term spectrum

try:
	frequencydomaingraph(pltltspec, "plot", "5. Rhythms and Rhythm Zones: Long Term Spectrum (FFT)", "LTS", fs/2.0, ltsf0, ltsmags, 0, 1.0, ltsmin, ltsmax, ltasfreqsplotcolor, 2.0)
	pltltspec.set_ylabel("Magnitude", fontsize=fontsize-1)
except:
	print("Long term signal spectrum graph error."); sys.exit()

#==================================================================
# ROW 2 - Long term envelope spectrum

try:
	frequencydomaingraph(pltaems, "plot", "6. Rhythms and Rhythm Zones: AEMS with main rhythm bar", "AEMS", fs/2.0, aemsf0, aemsmags, 0, 1.0, aemsmin, aemsmax, envfreqsplotcolor, 2.0)
	pltaems.set_ylabel("Magnitude", fontsize=fontsize-1)
except:
	print("AEMS graph error."); sys.exit()

#==================================================================
# ROW 3 - F0 spectrum

try:
	frequencydomaingraph(pltfems, "plot", "7. Rhythms and Rhythm Zones: FEMS with main rhythm bar", "FEMS", fs, femsf0, femsmags, 0, 1.0, femsmin, femsmax, ltfsfreqsplotcolor, 2.0)
	pltfems.set_ylabel("Magnitude", fontsize=fontsize-1)
except:
	print("FEMS graph error."); sys.exit()

#==================================================================
# ROW 4 - Descriptive text

try:
	plttext.set_title("8. Process Chains", fontsize=fontsize, fontweight='bold')
	plttext.text(0.01,0.1,infotext, fontsize=fontsize+1, fontweight='bold')
	plttext.set_xticks([])
	plttext.set_yticks([])
except:
	print("Information box error."); sys.exit()

#==================================================================
# GRAPH TERMINATION
#==================================================================

try:
	plt.tight_layout(pad=3, w_pad=-1, h_pad=1)
	plt.savefig(filetitle)
	plt.show()
	plt.close('all')
except:
	print("Graph termination error."); sys.exit()

#==================================================================
# Sound

if soundout:
	soundoutput(p, signal, volume)

#==============================================================


